(function (document, Joomla) {
  "use strict";

  var system_container;
  var options = Joomla.getOptions("miniteksystemmessages");
  var token = options.token + "=1";
  var is_site = options.is_site;
  var site_path =
    is_site != true ? options.site_path + "administrator/" : options.site_path;
  var lifetime = options.lifetime;
  var user_id = parseInt(options.user_id, 10);
  var application_messages = options.application_messages;
  var session_message = options.session_message;
  var session_redirect_link = options.session_redirect_link;
  var actions_log = options.actions_log;
  var logged_users = options.logged_users;
  var error_text = options.error_text;
  var success_text = options.success_text;
  var notice_text = options.notice_text;
  var warning_text = options.warning_text;

  var polipop;
  var appendTo = options.appendTo;
  var position = options.position;
  var layout = options.layout;
  var theme = options.theme;
  var icons = options.icons;
  var insert = options.insert;
  var spacing = options.spacing;
  var pool = options.pool;
  var sticky = options.sticky;
  var life = options.life;
  var progressbar = options.progressbar;
  var pauseOnHover = options.pauseOnHover;
  var headerText = options.headerText;
  var closer = options.closer;
  var closeText = options.closeText;
  var loadMoreText = options.loadMoreText;
  var hideEmpty = options.hideEmpty;
  var effect = options.effect;
  var easing = options.easing;
  var effectDuration = options.effectDuration;
  var group_messages = options.group_messages;

  // Create a new message
  function createMessage(content, type, isSticky) {
    var title;
    var _type = "notice";
    var _sticky = isSticky ? true : sticky;

    if (
      type === "error" ||
      type === "danger" ||
      type === "error alert-danger"
    ) {
      _type = "error";
      title = error_text;
    } else if (type === "message" || type === "success") {
      _type = "success";
      title = success_text;
    } else if (type === "warning") {
      _type = "warning";
      title = warning_text;
    } else {
      _type = "notice";
      title = notice_text;
    }

    // Create message
    polipop.add({
      type: _type,
      title: title,
      content: content,
      sticky: _sticky,
    });
  }

  // Check for expired user session
  function checkSession() {
    Joomla.request({
      url:
        site_path +
        "index.php?group=system&plugin=miniteksystemmessages&type=checkSession&" +
        token,
      method: "GET",
      onSuccess: (response, xhr) => {
        if (response) {
          var messages = JSON.parse(response);

          messages.forEach(function (element) {
            var content = "<div>" + element.message + "</div>";
            createMessage(content, element.type, true);
          });

          if (is_site && session_redirect_link)
            window.location.href = session_redirect_link;
        } else {
          var _lifetime = lifetime * 60000 + 1000;

          setTimeout(function () {
            checkSession();
          }, parseInt(_lifetime, 10));
        }
      },
      onError: (xhr) => {
        console.log(xhr);
      },
    });
  }

  // Create an observer instance - Observe the system-message-container for newly added messages
  function observeContainer() {
    if (!system_container) return;

    var observer = new MutationObserver(function (mutations) {
      mutations.forEach(function (mutation) {
        var newNodes = mutation.addedNodes;

        if (newNodes !== null) {
          newNodes.forEach(function (node) {
            var content = "";
            var _sticky;

            if (node.classList.contains("alert")) {
              var nodeClass = node.className;
              var type = nodeClass.substr(nodeClass.indexOf("-") + 1);

              if (type === "error alert-joomlaupdate") {
                type = "notice";
                var childrenDiv = [node];
                _sticky = true;
              } else {
                var childrenDiv = node.querySelectorAll(":scope > div");
              }

              // Group children
              if (group_messages === true) {
                childrenDiv.forEach(function (child) {
                  content += "<div>" + child.innerHTML + "</div>";
                });
                createMessage(content, type, _sticky);
              } else {
                childrenDiv.forEach(function (child) {
                  content = "<div>" + child.innerHTML + "</div>";
                  createMessage(content, type, _sticky);
                });
              }

              system_container.style.display = "none";
              system_container.innerHTML = "";
            }
          });
        }
      });
    });

    // Configuration of the observer
    var config = {
      attributes: true,
      childList: true,
      characterData: true,
    };

    // Pass in the target node, as well as the observer options
    observer.observe(system_container, config);
  }

  // Create new server-sent event
  function serverSentEvent() {
    var es = new EventSource(
      site_path +
        "index.php?group=system&plugin=miniteksystemmessages&type=serverSentEvent&" +
        token
    );

    var listener = function (event) {
      if (typeof event.data !== "undefined") {
        var messages = JSON.parse(event.data).messages;

        messages.forEach(function (message) {
          var content = '<div class="alert-message">' + message.text + "</div>";
          var type = message.type;

          createMessage(content, type);
        });
      }
    };

    es.addEventListener("open", listener);
    es.addEventListener("message", listener);
    es.addEventListener("error", listener);

    return es;
  }

  document.addEventListener("DOMContentLoaded", function () {
    system_container = document.querySelector("#system-message-container");

    // Polipop options
    var polipopOptions = {
      appendTo: appendTo,
      position: position,
      layout: layout,
      theme: theme,
      icons: icons,
      insert: insert,
      spacing: spacing,
      pool: pool,
      sticky: sticky,
      life: life,
      progressbar: progressbar,
      pauseOnHover: pauseOnHover,
      headerText: headerText,
      closer: closer,
      closeText: closeText,
      loadMoreText: loadMoreText,
      hideEmpty: hideEmpty,
      effect: effect,
      easing: easing,
      effectDuration: effectDuration,
    };

    // Initialize Polipop
    polipop = new Polipop("polipop", polipopOptions);

    // Parse application messages
    if (application_messages.length) {
      system_container.style.display = "none";
      system_container.innerHTML = "";

      application_messages.forEach(function (element) {
        createMessage(element.message, element.type);
      });
    }

    // Session expiration message
    if (session_message && user_id > 0) checkSession();

    // Observe container
    observeContainer();

    // Create event
    if (actions_log || logged_users) var es = serverSentEvent();

    if (typeof es !== "undefined") {
      window.addEventListener("beforeunload", function (event) {
        es.close();
      });
    }
  });
})(document, Joomla);
